/**********************************************************/
/*
    * Copyright (C) 2006 Lee Wilmot <dellfand.lee@dinglisch.net>

    This file is part of dellfand.

    dellfand is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    dellfand is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You can find a copy of the GNU General Public License in the
    file gpl.txt distributed with this file.
*/
/*************************************************************/

#if defined (__SVR4) && defined (__sun)
#include <sys/types.h>
#include <sys/sysi86.h>
#include <ia32/sys/psw.h>
#else
#include <sys/io.h>
#endif

#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <signal.h>
#include <unistd.h>

// Apparently necessary in some environments to avoid
// compile problems, thanks to Andreas Jackisch
#include <errno.h>

// Return values for error conditions

#define SYSTEM_ERROR 1
#define ARG_ERROR 2
#define FORK_ERROR 3
#define SID_ERROR 4
#define CHDIR_ERROR 5

#define VERSION 0.9

enum mode_type { foreground = 0, background, info };

// Globals

// Mode to run in, set according to command line arguments
// Global because needed by in interrupt handler, thanks
// to Markus Ethen for pointing this out.

mode_type modeG = info;

// Include code written by bofr@irq.org --------//
#include "fan.cc"

int check_args( float sleep_seconds, int off, int low, int high )
{
   if ( sleep_seconds < 0.2 ) {
	   puts(" <sleep_seconds> must be >= 0.2 "); 
	   return ARG_ERROR; 
   }
   else if( off < 25 || off > 60 ) { 
	   puts(" <off> must be in range 25..60 "); 
	   return ARG_ERROR; 
   }
   else if ( low < off || low > 60 ) { 
	   puts(" <low> must be in range <off>..60 "); 
	   return ARG_ERROR;
   } 
   else if ( high < low || high > 70 ) { 
	   puts(" <high> must be in range <low>..70) "); 
	   return ARG_ERROR;
   }
   else if ( modeG != background && modeG != foreground ) {
	   puts( "<mode> must be in range 0..1" ); 
	   return ARG_ERROR;
   }
   return 0;
}


int do_daemon_setup( void )
{
	umask(0);
         
	pid_t sid = setsid();
	if (sid < 0)
	       return SID_ERROR;

	if ( (chdir("/")) < 0 )
		return CHDIR_ERROR;

	close(STDIN_FILENO);
	close(STDOUT_FILENO);
	close(STDERR_FILENO);

	return 0;
}
int set_ioperms( void )
{
#if defined (__SVR4) && defined (__sun)
	int err = sysi86(SI86V86, V86SC_IOPL, PS_IOPL);
#else
	int err = ioperm(0x84,1,1);
#endif
	if ( err < 0 ) {
		puts( "error: syscall ioperm failed: probably you're not running as root." );
		return SYSTEM_ERROR;
	}

#if !(defined (__SVR4) && defined (__sun))
	if ( ( err = ioperm(0xb2,1,1) ) != 0 ) {
		puts( "Error: syscall ioperm failed: odd that." );
		return SYSTEM_ERROR;
	}
#endif

	return 0;
}

void do_main_loop( float sleep_seconds, int off, int low, int high )
{
	int done = 0;
	int last_set_status = -1;

	while ( done == 0 ) {
		int old_status = get_fan_status(0);
		int speed = get_fan_speed(0);
		int temp = get_cpu_temp();

		// Print data in info mode
		if ( modeG == info ) {
			printf( "Fan 0 Status %d Speed %d CPU Temp %dC\n", old_status, speed, temp );

			done = 1;
		}
		// Amend fan if not mode 2
		else {
			int desired_status = old_status;
	
   			if ( temp <= off ) 
				desired_status = 0;
			else if ( temp >= low && temp < high) 
				desired_status = 1;
			else if ( temp >= high )
				desired_status = 2;

			// Little check it got/stayed set to what we asked last time

			if ( 
				( last_set_status >= 0 ) &&
				( last_set_status != old_status )
			) {
				fprintf( stderr, "dellfand: warning: set fan 0 status to %d last cycle, it's now %d (BIOS interference ?)\n", last_set_status, old_status );
				last_set_status = -2;
			}

			// Only set fan if it's not what we already want (duh...)
			//
			if ( old_status != desired_status )
				set_fan( 0, desired_status );

			printf( "Fan 0 Status %d->%d Speed %d CPU Temp %dC\n", old_status, desired_status, speed, temp );

			// Record status, unless we've warned once already

			if ( last_set_status != -2 )
				last_set_status = desired_status;

  			usleep( (long) ( sleep_seconds * (float) 1000000 ) );
		}
	}
}

void prepare_to_exit( void )
{
	// If fan is off, set it to low since it will
	// stay off and laptop will overheat
	// If it was low or high, leave as is
	
	if ( get_fan_status(0) == 0 )
		set_fan( 0, 1 );
}


void int_handler( int sig )
{
	// Never adjust fan if we only wanted stats
	if ( modeG != 2 )
		prepare_to_exit();

	exit( 0 );
}

// 0 = fore
// 1 = back
// 2 = info
//
int main( int argn, char **args )
{
	// register int handler
	// we want to leave the fan on low when we
	// get INT in case it was an accidental term
	
	(void) signal( SIGINT, int_handler );

	// Process args
	
	float sleep_seconds;
	int off, low, high;

	if ( argn == 1 ) {
        	// Don't need set mode here for info, it's the default
		printf("v%.1f: ", VERSION ); 
	}
	else if ( argn == 6 ) {
		modeG = (mode_type) atoi(args[1]);
		sscanf( args[2], "%f", &sleep_seconds );
		off = atoi(args[3]);
		low = atoi(args[4]);
		high = atoi(args[5]);
	}
	else { 
		puts(" usage: dellfand [<mode> <sleep seconds> <off> <low> <high>]"); 
		return ARG_ERROR;
	}

	// Check args (if there were any)
  
	int err;
	if ( modeG != info ) {
		err = check_args( sleep_seconds, off, low, high );
		if ( err != 0 ) 
			return err;
	}
   
	// ioperms

	err = set_ioperms();
	if ( err != 0 )
		return err;

	pid_t pid = 0;

	// Fork if requested
	if ( modeG == background )
		pid = fork();
  
	// Fork failure
	if ( pid < 0 )
	   return FORK_ERROR;

	// Child, or 'parent' if no fork requested
	else if ( pid == 0 ) {

		if ( modeG == background ) {
			err = do_daemon_setup();
			if ( err != 0 )
				return err;
		}

		do_main_loop( sleep_seconds, off, low, high );
	}

	if ( modeG != info )
		prepare_to_exit();

	return 0;
}
